package pltag.german;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;


public abstract class AbstractGrammarExtractor {

	protected ArrayList<LexEntry> currLexicon;
	protected TigerTree currTigerTree;

	protected boolean multiAnchoredTrees = false;
	
	protected LexEntryComparator comparator = new LexEntryComparator();
	
	public AbstractGrammarExtractor() {
		currLexicon = new ArrayList<LexEntry>();
		currTigerTree = null;
	}

	public AbstractGrammarExtractor(boolean ma) {
		this();
		multiAnchoredTrees = ma;
	}
	
	
	public ArrayList<LexEntry> createLexicon(TigerTree tree) {
		currLexicon.clear();
		currTigerTree = tree;
		
		extract(tree.getRoot(), new LexEntry(tree.getRoot(), multiAnchoredTrees));
		
		// sort the lexicon entries according to the nodeID of their leafLink
		Collections.sort(currLexicon, comparator);
		return currLexicon;
	}
	
	abstract protected void extract(TreeNode root, LexEntry lexEntry);
	

	class LexEntryComparator implements Comparator<LexEntry> {
		public int compare(LexEntry l1, LexEntry l2) {
			int r1 = l1.getLeafLink().getID();
			int r2 = l2.getLeafLink().getID();
			if (r1 > r2) {
				return 1;
			}
			else if (r1 < r2) {
				return -1;
			}
			else {
				return 0;
			}
		}
	}
}
