package pltag.german;

import java.util.ArrayList;
import java.util.HashMap;


public class CompositeTree extends ElementaryTree {
	
	private ArrayList<ElementaryTree> notIntegratedYet = new ArrayList<ElementaryTree>();
	
	private ArrayList<ElementaryTree> individualTrees = new ArrayList<ElementaryTree>();

	/*
	 * Creates a "copy" of the elementary tree as the start of the composite tree.
	 * The trace set and the lexEntry list are not copied at the moment
	 */
	public CompositeTree(ElementaryTree elementaryTree) {
		setType(elementaryTree.getType());
		ElementaryTreeNode root = makeCopyAndAdd(elementaryTree.getRoot(), elementaryTree);
		setRoot(root);
		if (getType() == NodeStatus.MOD) {
			setModParent(elementaryTree.getModParent());
		}
		individualTrees.add(elementaryTree);
	}
	
	public void tryToIntegrate(ElementaryTree eTree, StringBuffer errorString) {
		tryToIntegrate(eTree, errorString, true, false);
	}

	public void tryToIntegrate(ElementaryTree eTree, StringBuffer errorString, boolean sisterAdjOnlyOnFullNodes, boolean prediction) {
		boolean success = integrate(eTree, errorString, sisterAdjOnlyOnFullNodes, prediction);
		individualTrees.add(eTree);
		if (!success) {
			notIntegratedYet.add(eTree);
		}
	}
	
	public void tryToIntegratePrediction(ElementaryTree eTree, StringBuffer errorString, boolean sisterAdjOnlyOnFullNodes) {
		tryToIntegrate(eTree, errorString, sisterAdjOnlyOnFullNodes, true);
	}
	
	public ArrayList<ElementaryTree> getIndividualTrees() {
		return individualTrees;
	}
	
	public boolean tryMore(StringBuffer errorString) {
		return tryMore(errorString, true, false);
	}
	
	public boolean tryMorePrediction(StringBuffer errorString, boolean sisterAdjOnlyOnFullNodes) {
		return tryMore(errorString, sisterAdjOnlyOnFullNodes, true);
	}
	
	public boolean tryMore(StringBuffer errorString, boolean sisterAdjOnlyOnFullNodes, boolean prediction) {
		boolean somethingGoesOn = true;
		while (notIntegratedYet.size() > 0 && somethingGoesOn) {
			int i = 0;
			ElementaryTree eTree = null;
			while (i < notIntegratedYet.size()) {
				eTree = notIntegratedYet.get(i);
				boolean success = integrate(eTree, errorString, sisterAdjOnlyOnFullNodes, prediction);
				if (success) {
					break;
				}
				i++;
			}
			if (i < notIntegratedYet.size()) {
				// eTree i was integrated
				notIntegratedYet.remove(i);
			}
			else {
				somethingGoesOn = false;
			}
		}
		if (notIntegratedYet.size() == 0) {
			return true;
		}
		else {
			return false;
		}
	}
	
	private boolean integrate(ElementaryTree eTree, StringBuffer errorString, boolean sisterAdjOnlyOnFullNodes, boolean prediction) {
		boolean integrated = false;
		StringBuffer localErrorString = new StringBuffer();
		if (!integrated && eTree.getType() == NodeStatus.ARG) {
			// Try to substitute eTree into the current tree
			ElementaryTreeNode node = this.getNodeWithBottom(eTree.getRoot().getTreeNodeBottom());
							//this.getNodeWithTopAndBottom(eTree.getRoot().getTreeNodeTop(), 
							//eTree.getRoot().getTreeNodeBottom());//containsID(eTree.getRoot().getID());
			if (node != null && node.isSubst()) {
				if ( node.getChildren().size() == 0 
						&& labelsMatch(node, eTree.getRoot())) {
					ElementaryTreeNode parent = node.getParent();
					// replace this substitution site with the initial eTree
					int i = parent.getChildren().indexOf(node);
					getNodes().remove(node.getID());
					ElementaryTreeNode rootCopy = makeCopyAndAdd(eTree.getRoot(), eTree);
					// keep the "top" part of the substitution node
					rootCopy.setFunctionalCategory(node.getFunctionalCategory());
					rootCopy.setOriginTop(node.getOriginTop());
					rootCopy.setTreeNodeTop(node.getTreeNodeTop());
					rootCopy.setID(node.getID());
					parent.setChild(i, rootCopy);
					rootCopy.setParent(parent);
					// "unify" morphology
					MorphInfo m = node.getMorphology().unify(eTree.getRoot().getMorphology());
					if (m == null) {
						localErrorString.append("### Clash in Morphology Features: " + node + ": " + eTree);
					}
					else {
						rootCopy.setMorphology(m);
					}
					integrated = true;
				}
				else if (!sisterAdjOnlyOnFullNodes && node.getChildren().size() > 0
						&& labelsMatch(node, eTree.getRoot())) {
					// this is a special case if the flag sisterAdjOnlyOnFullNodes is set to false.
					// sister-adjunction of modifier trees can then also happen to nodes of which we have
					// not yet seen the bottom part, i.e. open substitution nodes.
					// the node will keep its substitution status, but it already has children.
					// we need to find out, where exactly we put our initial eTree
					// TODO
					System.out.println("##### Error: Not implemented yet!!!");
				}
				else {
					localErrorString.append("### Cannot substitute here: " + node + ": " + eTree);
				}
			}
		}
		if (!integrated && getType() == NodeStatus.ARG) {
			// Try to substitute the current tree into the eTree
			ElementaryTreeNode node = eTree.getNodeWithBottom(getRoot().getTreeNodeBottom());
							//eTree.getNodeWithTopAndBottom(getRoot().getTreeNodeTop(), 
							//getRoot().getTreeNodeBottom());//containsID(getRoot().getID());
			if (node != null && node.isSubst()) {
				if (node.getChildren().size() == 0 
						&& labelsMatch(node, getRoot())) {
					
					ElementaryTree treeCopy = makeCopy(eTree);
					ElementaryTreeNode node2 = treeCopy.getNodeWithTopAndBottom(getRoot().getTreeNodeTop(), 
							getRoot().getTreeNodeBottom());//containsID(getRoot().getID());					
					ElementaryTreeNode parent = node2.getParent();
					// replace this substitution site with the current tree
					int i = parent.getChildren().indexOf(node2);
					parent.setChild(i, getRoot());
					getRoot().setParent(parent);
					// keep the "top" part of the substitution node
					getRoot().setFunctionalCategory(node2.getFunctionalCategory());
					getRoot().setOriginTop(node2.getOriginTop());
					getRoot().setTreeNodeTop(node2.getTreeNodeTop());
					// "unify" morphology
					MorphInfo m = node2.getMorphology().unify(getRoot().getMorphology());
					if (m == null) {
						localErrorString.append("### Clash in Morphology Features: " + node2);
					}
					else {
						getRoot().setMorphology(m);
					}
					// fix the current tree, add information from the eTree
					addNodes(treeCopy.getNodes());
					setType(treeCopy.getType());
					setRoot(treeCopy.getRoot());
					if (getType() == NodeStatus.ADJ) {
						setFoot(treeCopy.getFoot());
					}
					if (getType() == NodeStatus.MOD) {
						setModParent(treeCopy.getModParent());
					}
					integrated = true;
				}
				else {
					localErrorString.append("### Cannot substitute here: " + node);
				}
			}
		}
		if (!integrated && eTree.getType() == NodeStatus.ADJ) {
			// Try to adjoin eTree into the current tree
			ElementaryTreeNode node = this.getNodeWithTop(eTree.getRoot().getTreeNodeBottom());
			if (node != null) {
				if (eTree.getFoot() != null && eTree.getFoot().getChildren().size() == 0 
						&& labelsMatch(eTree.getRoot(), eTree.getFoot())
						&& labelsMatch(eTree.getRoot(), node)) {
					// main work factored out
					adjoinETreeIntoCurrentTree(eTree, node);
					integrated = true;
				}
				else {
					localErrorString.append("### Something is wrong with adjunction: " + 
									eTree.toStringStruct(new HashMap<String, Trace>()));
				}
			}
			if (!integrated) {
				// it might be that the auxiliary tree cannot directly adjoin. there might be
				// other (intervening) auxiliary trees. but we might still be able
				// to adjoin the current one.
				TreeNode bot = eTree.getRoot().getTreeNodeBottom();
				while (!integrated && bot.getTop() != null && bot.getTop().isAdjTree() && bot == bot.getTop().getFoot()) {
					TreeNode rootOfInterveningAdj = bot.getTop().getRoot();
					ElementaryTreeNode node2 = this.getNodeWithTop(rootOfInterveningAdj);
					if (node2 != null) {
						if (eTree.getFoot() != null && eTree.getFoot().getChildren().size() == 0 
								&& labelsMatch(eTree.getRoot(), eTree.getFoot()) 
								&& labelsMatch(eTree.getRoot(), node2)) {
							adjoinETreeIntoCurrentTree(eTree, node2);
							integrated = true;
						}
						else {
							localErrorString.append("### Something is wrong with adjunction: " + 
											eTree.toStringStruct(new HashMap<String, Trace>()));
						}
						break;
					}
					else {
						bot = rootOfInterveningAdj;
					}
				}
			}
		}
		
		if (!integrated && getType() == NodeStatus.ADJ) {
			// Try to adjoin the current tree into the eTree
			ElementaryTreeNode node = eTree.getNodeWithTop(getRoot().getTreeNodeBottom());
			if (node != null) {
				if (getFoot() != null && getFoot().getChildren().size() == 0 
						&& labelsMatch(getRoot(), getFoot())
						&& labelsMatch(getRoot(), node)) {
					ElementaryTree treeCopy = makeCopy(eTree);
					addNodes(treeCopy.getNodes());
					ElementaryTreeNode node2 = treeCopy.getNodeWithTop(getRoot().getTreeNodeBottom());
					// main work factored out
					adjoinCurrentTreeIntoETree(treeCopy, node2);
					integrated = true;
				}
				else {
					localErrorString.append("### Something is wrong with adjunction: " 
									+ toStringStruct(new HashMap<String, Trace>()));
				}
			}
			if (!integrated) {
				// it might be that the auxiliary tree cannot directly adjoin. there might be
				// other (intervening) auxiliary trees. but we might still be able
				// to adjoin the current one.
				TreeNode bot = getRoot().getTreeNodeBottom();
				while (!integrated && bot.getTop() != null && bot.getTop().isAdjTree() && bot == bot.getTop().getFoot()) {
					TreeNode rootOfInterveningAdj = bot.getTop().getRoot();
					ElementaryTreeNode node2 = eTree.getNodeWithTop(rootOfInterveningAdj);
					if (node2 != null) {
						if (getFoot() != null && getFoot().getChildren().size() == 0 
								&& labelsMatch(getRoot(), getFoot()) 
								&& labelsMatch(getRoot(), node2)) {
							ElementaryTree treeCopy = makeCopy(eTree);
							addNodes(treeCopy.getNodes());
							ElementaryTreeNode node3 = treeCopy.getNodeWithTop(rootOfInterveningAdj);
							// main work factored out
							adjoinCurrentTreeIntoETree(treeCopy, node3);
							integrated = true;
						}
						else {
							localErrorString.append("### Something is wrong with adjunction: " + 
													toStringStruct(new HashMap<String, Trace>()));
						}
						break;
					}
					else {
						bot = rootOfInterveningAdj;
					}
				}
			}
		}
		
		
		
		if (!integrated && eTree.getType() == NodeStatus.MOD) {
			// Try to sister-adjoin the eTree into the current tree.
			// We have stored the "correct" mother node for sister-adjunction,
			// but we can only sister-adjoin if the direct left or right sibling
			// is already in the current tree
			ElementaryTreeNode node = this.getNodeWithBottom(eTree.getModParent().getID());
			if (node != null) {
				int eTreeID = eTree.getRoot().getID();
				int leftSibID = -999;
				int rightSibID = -999;
				int prev = -999;
				for (int i = 0; i < eTree.getModParent().getChildren().size(); i++) {
					TreeNode originalChild = eTree.getModParent().getChildren().get(i);
					if (originalChild.getID() == eTreeID) {
						leftSibID = prev;
						if (i < eTree.getModParent().getChildren().size()-1) {
							rightSibID = eTree.getModParent().getChildren().get(i+1).getID();
						}
						break;
					}
					prev = originalChild.getID();
				}
				if (!integrated && leftSibID != -999) {
					ElementaryTreeNode leftSib = containsID(leftSibID);
					if (leftSib != null) {
						// the left sibling is already in the current tree,
						// we can sister-adjoin the eTree right of it
						int i = node.getIndexOfChildWithID(leftSibID);
						ElementaryTree treeCopy = makeCopy(eTree);
						addNodes(treeCopy.getNodes());
						node.addChild(i+1, treeCopy.getRoot());
						treeCopy.getRoot().setParent(node);
						integrated = true;
					}
				}
				if (!integrated && rightSibID != -999) {
					ElementaryTreeNode rightSib = containsID(rightSibID);
					if (rightSib != null) {
						// the right sibling is already in the current tree,
						// we can sister-adjoin the eTree left of it
						int i = node.getIndexOfChildWithID(rightSibID);
						ElementaryTree treeCopy = makeCopy(eTree);
						addNodes(treeCopy.getNodes());
						node.addChild(i, treeCopy.getRoot());
						treeCopy.getRoot().setParent(node);
						integrated = true;
					}
				}
				if (!integrated) {
					// no direct sibling is already in the tree.
					// try to search for the next best one from the left,
					// then the modifier can be put to its left
					ElementaryTreeNode nextRight = null;
					boolean found = false;
					for (int i = 0; i < eTree.getModParent().getChildren().size(); i++) {
						TreeNode originalChild = eTree.getModParent().getChildren().get(i);
						if (originalChild.getID() == eTreeID) {
							found = true;
						}
						else if (found) {
							nextRight = containsID(originalChild.getID());
							if (nextRight != null) {
								break;
							}
						}
					}
					if (nextRight != null) {
						int i = node.getIndexOfChildWithID(nextRight.getID());
						ElementaryTree treeCopy = makeCopy(eTree);
						addNodes(treeCopy.getNodes());
						node.addChild(i, treeCopy.getRoot());
						treeCopy.getRoot().setParent(node);
						integrated = true;
					}

				}
				if (!integrated) {
					// no direct sibling is already in the tree.
					// try to search for the next best one from the right,
					// then the modifier can be put to its right
					ElementaryTreeNode nextLeft = null;
					boolean found = false;
					for (int i = eTree.getModParent().getChildren().size()-1; i >= 0 ; i--) {
						TreeNode originalChild = eTree.getModParent().getChildren().get(i);
						if (originalChild.getID() == eTreeID) {
							found = true;
						}
						else if (found) {
							nextLeft = containsID(originalChild.getID());
							if (nextLeft != null) {
								break;
							}
						}
					}
					if (nextLeft != null) {
						int i = node.getIndexOfChildWithID(nextLeft.getID());
						ElementaryTree treeCopy = makeCopy(eTree);
						addNodes(treeCopy.getNodes());
						node.addChild(i+1, treeCopy.getRoot());
						treeCopy.getRoot().setParent(node);
						integrated = true;
					}

				}
				
				if (!integrated && prediction && node.getChildren().size() == 0) {
					// this happens when we combine (cut) prediction trees with each other
					ElementaryTree treeCopy = makeCopy(eTree);
					addNodes(treeCopy.getNodes());
					node.addChild(0, treeCopy.getRoot());
					treeCopy.getRoot().setParent(node);
					integrated = true;
				}
				
				if (!integrated && !sisterAdjOnlyOnFullNodes) {
					// the modifier cannot be integrated with the partial tree here,
					// because it wants to sister-adjoin to a substitution node for which
					// we have not seen the initial tree yet.
					// if the condition sisterAdjOnlyOnFullNodes == false, this configuration
					// is possible.
					// we make the modifier tree, the child of the substitution node,
					// but keep its substitution status. the substitution operation above
					// will also be modified such that this kind of operation is supported
					if (node.isSubst() && node.getChildren().size() == 0) {
						ElementaryTree treeCopy = makeCopy(eTree);
						addNodes(treeCopy.getNodes());
						node.addChild(treeCopy.getRoot());
						treeCopy.getRoot().setParent(node);
						integrated = true;
					}
				}
			}
			// it seems that we try to integrate the eTree at a node that 
			// in the training data would be separated (or completed, depending on how we are looking at it)
			// by an auxiliary tree.
			// with the sisterAdjOnlyOnFullNodes= false setting, we are allowed to
			// sister-adjoin the eTree into the current tree if we have seen only the top part. 
			// NOTE: if the aux tree adjoins later, the modifier tree will have a position at its foot,
			// i.e. not at the root as it was in the original tree.
			// an additional model/mechanism would be needed that allows to move modifiers at foot nodes up 
			// to the root node.
			if (!integrated && !sisterAdjOnlyOnFullNodes) {
				node = this.getNodeWithTop(eTree.getModParent().getID());
				if (node != null) {
					if (eTree.getModParent().getBottom().isAdjTree()) {
						// find out whether eTree is right or left of the spine in
						// the original tree
						int eTreeID = eTree.getRoot().getID();
						int foundETree = -1;
						int foundSpine = -1;
						int d = 0;
						for (int i = 0; i < eTree.getModParent().getChildren().size(); i++) {
							TreeNode originalChild = eTree.getModParent().getChildren().get(i);
							if (originalChild.getID() == eTreeID) {
								foundETree = i;
							}
							if (eTree.getModParent().getBottom().getElementaryTree().isOnSpine(originalChild.getID())) { 
								foundSpine = i;
							}
							if (foundETree > -1 && foundSpine > -1) {
								if (foundETree < foundSpine) d = -1;
								else d = 1;
								break;
							}
						}
						if (d != 0) {
							ElementaryTree treeCopy = makeCopy(eTree);
							addNodes(treeCopy.getNodes());
							if (d == -1) {
								node.addChild(0, treeCopy.getRoot());
							}
							else if (d == 1) {
								node.addChild(node.getChildren().size(), treeCopy.getRoot());
							}
							treeCopy.getRoot().setParent(node);
							integrated = true;
						}
					}
					else {
						localErrorString.append("### Something is wrong with sister-adjunction (sisterAdjOnlyOnFullNodes=" +
									sisterAdjOnlyOnFullNodes + "): " + toStringStruct(new HashMap<String, Trace>()));
					}
				}
			}
			// this is a follow-up of the previous case:
			// the node to which we want to sister-adjoin might be made up of the root of an aux tree
			// (BOT) and the foot of another aux tree (TOP).
			// we go up the tree and try to sister-adjoin to the root of the TOP aux tree.
			if (!integrated && !sisterAdjOnlyOnFullNodes) {
				TreeNode treeNode = eTree.getModParent();
				while (treeNode != null && treeNode.getTop().isAdjTree()) {
					node = this.getNodeWithTop(treeNode.getTop().getRoot().getID());
					if (node != null) {
						// this is a very special case that occurs during incremental parsing.
						// the modifier will always attach as the left-most child
						// (if it was right-most, this case would not occur)
						// TODO: correct?!?!
						ElementaryTree treeCopy = makeCopy(eTree);
						addNodes(treeCopy.getNodes());
						node.addChild(0, treeCopy.getRoot());
						treeCopy.getRoot().setParent(node);
						integrated = true;
						break;
					}
					else {
						treeNode = treeNode.getTop().getRoot();
					}
				}
			}
		}
		
		if (!integrated && getType() == NodeStatus.MOD) {
			// Try to sister-adjoin the current tree into the eTree.
			// We have stored the "correct" mother node for sister-adjunction,
			// but we can only sister-adjoin if the direct left or right sibling
			// is already in the eTree
			ElementaryTreeNode node = eTree.getNodeWithBottom(getModParent().getID());
			if (node != null) {
				int modTreeID = getRoot().getID();
				int leftSibID = -999;
				int rightSibID = -999;
				int prev = -999;
				for (int i = 0; i < getModParent().getChildren().size(); i++) {
					TreeNode originalChild = getModParent().getChildren().get(i);
					if (originalChild.getID() == modTreeID) {
						leftSibID = prev;
						if (i < getModParent().getChildren().size()-1) {
							rightSibID = getModParent().getChildren().get(i+1).getID();
						}
						break;
					}
					prev = originalChild.getID();
				}
				if (!integrated && leftSibID != -999) {
					ElementaryTreeNode leftSib = eTree.containsID(leftSibID);
					if (leftSib != null) {
						// the left sibling is already in the eTree,
						// we can sister-adjoin the current tree right of it
						ElementaryTree treeCopy = makeCopy(eTree);
						ElementaryTreeNode node2 = treeCopy.getNodeWithBottom(getModParent().getID());
						//ElementaryTreeNode leftSib2 = treeCopy.containsID(leftSibID);
						int i = node2.getIndexOfChildWithID(leftSibID);
						addNodes(treeCopy.getNodes());
						node2.addChild(i+1, getRoot());
						getRoot().setParent(node2);
						// fix stuff
						setRoot(treeCopy.getRoot());
						setType(treeCopy.getType());
						if (getType() == NodeStatus.ADJ) {
							setFoot(treeCopy.getFoot());
						}
						else {
							setFoot(null);
						}
						if (getType() == NodeStatus.MOD) {
							setModParent(treeCopy.getModParent());
						}
						else {
							setModParent(null);
						}
						integrated = true;
					}
				}
				if (!integrated && rightSibID != -999) {
					ElementaryTreeNode rightSib = eTree.containsID(rightSibID);
					if (rightSib != null) {
						// the right sibling is already in the eTree,
						// we can sister-adjoin the current tree left of it
						ElementaryTree treeCopy = makeCopy(eTree);
						ElementaryTreeNode node2 = treeCopy.getNodeWithBottom(getModParent().getID());
						//ElementaryTreeNode rightSib2 = treeCopy.containsID(rightSibID);
						int i = node2.getIndexOfChildWithID(rightSibID);
						addNodes(treeCopy.getNodes());
						node2.addChild(i, getRoot());
						getRoot().setParent(node2);
						// fix stuff
						setRoot(treeCopy.getRoot());
						setType(treeCopy.getType());
						if (getType() == NodeStatus.ADJ) {
							setFoot(treeCopy.getFoot());
						}
						else {
							setFoot(null);
						}
						if (getType() == NodeStatus.MOD) {
							setModParent(treeCopy.getModParent());
						}
						else {
							setModParent(null);
						}
						integrated = true;
					}
				}
			}
		}
		
		if (!integrated) {
			errorString.append(localErrorString);
		}
		
		return integrated;
	}
	
	private void adjoinCurrentTreeIntoETree(ElementaryTree treeCopy, ElementaryTreeNode node2) {
		ArrayList<ElementaryTreeNode> childrenOfNode2 = node2.getChildren();
		node2.removeAllChildren();
		for (ElementaryTreeNode childOfAdj : getRoot().getChildren()) {
			node2.addChild(childOfAdj);
			childOfAdj.setParent(node2);
		}
		// "unify" morphology
		MorphInfo m = node2.getMorphology().unify(getRoot().getMorphology());
		if (m == null) {
			//localErrorString.append("### Clash in Morphology Features: " + node2);
		}
		else {
			node2.setMorphology(m);
		}
		// foot
		getFoot().setTreeNodeBottom(node2.getTreeNodeBottom());
		node2.setTreeNodeBottom(getRoot().getTreeNodeBottom());
		for (ElementaryTreeNode childOfNode : childrenOfNode2) {
			getFoot().addChild(childOfNode);
			childOfNode.setParent(getFoot());
		}
		// "unify" morphology
		MorphInfo m2 = getFoot().getMorphology().unify(node2.getMorphology());
		if (m2 == null) {
			//localErrorString.append("### Clash in Morphology Features: " + node2);
		}
		else {
			getFoot().setMorphology(m2);
		}
		// origins
		ElementaryTree bot = node2.getOriginBottom();
		node2.setOriginBottom(getRoot().getOriginBottom());
		getFoot().setOriginBottom(bot);
		if (node2.isSubst()) {
			getFoot().setSubst();
			node2.setSubst(false);
		}
		setRoot(treeCopy.getRoot());
		setType(treeCopy.getType());
		if (getType() == NodeStatus.ADJ) {
			setFoot(treeCopy.getFoot());
		}
		else {
			setFoot(null);
		}
		if (getType() == NodeStatus.MOD) {
			setModParent(treeCopy.getModParent());
		}
		
	}

	private void adjoinETreeIntoCurrentTree(ElementaryTree eTree, ElementaryTreeNode node) {
		ElementaryTree treeCopy = makeCopy(eTree);
		addNodes(treeCopy.getNodes());
		// put correct "old" root
		getNodes().remove(treeCopy.getRoot().getID());
		addNode(node); // 
		
		ArrayList<ElementaryTreeNode> childrenOfNode = node.getChildren(); // empty if node is a pre-terminal, see below
		node.removeAllChildren();
		for (ElementaryTreeNode childOfAdj : treeCopy.getRoot().getChildren()) {
			node.addChild(childOfAdj);
			childOfAdj.setParent(node);
		}
		// "unify" morphology
		MorphInfo m = node.getMorphology().unify(eTree.getRoot().getMorphology());
		if (m == null) {
			//localErrorString.append("### Clash in Morphology Features: " + node2);
		}
		else {
			node.setMorphology(m);
		}
		
		treeCopy.getFoot().setTreeNodeBottom(node.getTreeNodeBottom());
		node.setTreeNodeBottom(eTree.getRoot().getTreeNodeBottom());
		for (ElementaryTreeNode childOfNode : childrenOfNode) {
			treeCopy.getFoot().addChild(childOfNode);
			childOfNode.setParent(treeCopy.getFoot());
		}
		// "unify" morphology
		MorphInfo m2 = treeCopy.getFoot().getMorphology().unify(node.getMorphology());
		if (m2 == null) {
			//localErrorString.append("### Clash in Morphology Features: " + node2);
		}
		else {
			treeCopy.getFoot().setMorphology(m2);
		}
		
		// special case: node is a pre-terminal
		if (node.isLeaf()) {
			treeCopy.getFoot().setLeaf();
			treeCopy.getFoot().setLeaf(node.getLeaf());
			treeCopy.getFoot().setLemma(node.getLemma());
			node.setLeaf("");
			node.setLemma("");
			node.setNotLeaf();
		}
		
		// origins
		ElementaryTree bot = node.getOriginBottom();
		node.setOriginBottom(treeCopy.getRoot().getOriginBottom());
		treeCopy.getFoot().setOriginBottom(bot);
		
		if (node.isSubst()) {
			treeCopy.getFoot().setSubst();
			node.setSubst(false);
		}
	}
	
	private ElementaryTreeNode makeCopyAndAdd(ElementaryTreeNode node, ElementaryTree eTree) {
		
		ElementaryTreeNode copy = new ElementaryTreeNode(node);
		
		if (eTree.getType() == NodeStatus.ADJ && node == eTree.getFoot()) {
			setFoot(copy);
		}
		
		for (ElementaryTreeNode child : node.getChildren()) {
			ElementaryTreeNode childCopy = makeCopyAndAdd(child, eTree);
			childCopy.setParent(copy);
			copy.addChild(childCopy);
		}
		
		addNode(copy);
		return copy;
	}
	
	private ElementaryTree makeCopy(ElementaryTree eTree) {
		ElementaryTree newETree = new ElementaryTree();
		newETree.setType(eTree.getType());
		makeCopyHelper(eTree.getRoot(), eTree, newETree);
		if (eTree.getModParent() != null) {
			newETree.setModParent(eTree.getModParent());
		}
		return newETree;
	}
	
	private ElementaryTreeNode makeCopyHelper(ElementaryTreeNode node, ElementaryTree eTree,
					ElementaryTree newETree) {
		
		ElementaryTreeNode copy = new ElementaryTreeNode(node);
		
		if (eTree.getType() == NodeStatus.ADJ && node == eTree.getFoot()) {
			newETree.setFoot(copy);
		}
		if (node == eTree.getRoot()) {
			newETree.setRoot(copy);
		}
		if (node == eTree.getAnchor()) {
			newETree.setAnchor(copy);
		}
		
		for (ElementaryTreeNode child : node.getChildren()) {
			ElementaryTreeNode childCopy = makeCopyHelper(child, eTree, newETree);
			childCopy.setParent(copy);
			copy.addChild(childCopy);
		}
		
		newETree.addNode(copy);
		return copy;
	}

	public boolean correspondsTo(TigerTree tiger, StringBuffer errorString) {
		return correspondance(getRoot(), tiger.getRoot(), errorString);
	}

	private boolean correspondance(ElementaryTreeNode node, TreeNode tigerNode,
			StringBuffer errorString) {
		if (node.getPhrasalCategory().equals(tigerNode.getPhrasalCategory()) &&
				node.getFunctionalCategory().equals(tigerNode.getFunctionalCategory())) {
			if ((node.isLeaf() && !tigerNode.isLeaf()) || (!node.isLeaf() && tigerNode.isLeaf())) {
				errorString.append("### Terminals don't match: " + node + " / " + tigerNode);
				return false;
			}
			
			if (node.getChildren().size() == tigerNode.getChildren().size()) {
				for (int i = 0; i < node.getChildren().size(); i++) {
					correspondance(node.getChildren().get(i), tigerNode.getChildren().get(i), errorString);
				}
			}
			else {
				errorString.append("### No. of children don't match: "+ node + " / " + tigerNode);
				return false;
			}
		}
		else {
			errorString.append("### Categories don't match: " + node + " / " + tigerNode);
			return false;
		}
		return true;
	}
	
	private boolean labelsMatch(ElementaryTreeNode n1, ElementaryTreeNode n2) {
		return 
				n1.getPhrasalCategory().equals(n2.getPhrasalCategory())
				&& n1.getMorphology().equals(n2.getMorphology());
	}

	
//	public String toString() {
//		StringBuffer b = new StringBuffer();
//		toStringStruct(getRoot(), b);
//		return b.toString();
//	}
	
	private void toStringStruct(ElementaryTreeNode node, StringBuffer sb) {
		int top = individualTrees.indexOf(node.getOriginTop());
		int bottom = individualTrees.indexOf(node.getOriginBottom());
		String topOrigin;
		String bottomOrigin;
		if (top == -1) {
			topOrigin = "null";
		}
		else {
			topOrigin = String.valueOf(top+1);
		}
		if (bottom == -1) {
			bottomOrigin = "null";
		}
		else {
			bottomOrigin = String.valueOf(bottom+1);
		}
		
		sb.append("(");
		sb.append(node.getPhrasalCategory());
		if (!node.getFunctionalCategory().equals("--")) {
			sb.append("-");
			sb.append(node.getFunctionalCategory());
		}
		
		if (node.isSubst()) {
			sb.append("^").append(topOrigin).append("_").append("null");
			sb.append("!");
		}
		else if (getType() == NodeStatus.MOD && node == getRoot()) {
			sb.append("^").append(topOrigin).append("_").append(bottomOrigin);
			sb.append("*");
		}
		else if (getType() == NodeStatus.ADJ && node == getFoot()) {
			sb.append("^").append(topOrigin).append("_").append("null");
			sb.append("*");
		}
		else if (node == getRoot()) {
			sb.append("^").append("null").append("_").append(bottomOrigin);
		}
		else {
			sb.append("^").append(topOrigin).append("_").append(bottomOrigin);
		}
		
		sb.append(" ");

		for (ElementaryTreeNode child : node.getChildren()) {
			toStringStruct(child, sb);
		}

		
		sb.append(")");
	}

}
