package pltag.german;

import java.util.ArrayList;

public class DerivationFinder {
	
	private LTAGDerivation currDeriv = null;
	private CompositeTree currTree = null;


	public LTAGDerivation computeDerivation(CompositeTree tree, ArrayList<ElementaryTree> lex) {
		currDeriv = new LTAGDerivation(lex);
		currTree = tree;
		findStart(tree.getRoot());
		computeDerivation(tree.getRoot());
		return currDeriv;
	}

	private void findStart(ElementaryTreeNode node) {
		ElementaryTree newBottom = node.getOriginBottom();
		while (newBottom.getType() != NodeStatus.ARG) {
			ElementaryTreeNode foot = currTree.containsID(newBottom.getFoot().getID());
			newBottom = foot.getOriginBottom();
		}
		currDeriv.setStart(newBottom);
	}

	private void computeDerivation(ElementaryTreeNode node) {
		ElementaryTree top = node.getOriginTop();
		ElementaryTree bottom = node.getOriginBottom();
		if (top != bottom) { // some operation took place here
			// adjunction
			if (bottom.getType() == NodeStatus.ADJ && bottom.getRoot().getID() == node.getID()) {
				ElementaryTreeNode foot = bottom.getFoot();
				ElementaryTree target = currTree.containsID(foot.getID()).getOriginBottom();
				String gorn = target.calcGornTreeNodeBottom(foot.getID());
				currDeriv.addRelation(new DerivationRel(bottom, target, gorn));
			}
			// substitution
			if (top != null && top.getNodeWithTop(node.getID()).isSubst()) {
				ElementaryTree newBottom = bottom;
				while (newBottom.getType() != NodeStatus.ARG) {
					ElementaryTreeNode foot = currTree.containsID(newBottom.getFoot().getID());
					newBottom = foot.getOriginBottom();
				}
				String gorn = top.calcGornTreeNodeTop(node.getID());
				currDeriv.addRelation(new DerivationRel(newBottom, top, gorn));
			}
			
		}
		// sister-adjunction
		if (top != null && top.getType() == NodeStatus.MOD && top.getRoot().getID() == node.getID()) {
			ElementaryTreeNode parent = node.getParent();
			ElementaryTree target = currTree.containsID(parent.getID()).getOriginBottom();
			String gorn = target.calcGornTreeNodeTop(parent.getID());
			currDeriv.addRelation(new DerivationRel(top, target, gorn));
		}
		
		//recursion
		for (ElementaryTreeNode child : node.getChildren()) {
			computeDerivation(child);
		}
	}
	
}
