package pltag.german;

import java.util.ArrayList;
import java.util.HashMap;


public class ElementaryTree {
	
	private NodeStatus type = NodeStatus.UNDEF;
	
	private ElementaryTreeNode root;
	private ElementaryTreeNode foot;
	private ElementaryTreeNode anchor;
	private TreeNode modParent;
	private HashMap<Integer, ElementaryTreeNode> nodes = new HashMap<Integer, ElementaryTreeNode>();
	
	private HashMap<String, ElementaryTreeNode[]> traceSet = new HashMap<String, ElementaryTreeNode[]>();
	protected HashMap<String, Integer> traceMappingForOutput = new HashMap<String, Integer>();
	
	// the lexEntries which lead to this elementary tree (at least one)
	private ArrayList<LexEntry> lexEntries = new ArrayList<LexEntry>();
	private boolean partOfMAT = false;
	private ArrayList<ElementaryTree> individuals = new ArrayList<ElementaryTree>();
	private ArrayList<ElementaryTreeNode> otherAnchors = new ArrayList<ElementaryTreeNode>();
	
	private boolean motherLevelForSisterAdj = false;
	private boolean modifier_auxiliary = false;
	
	private MorphAnnotator morphAnnotator = new MorphAnnotator();
	
	public ElementaryTree() {
		
	}

	public ElementaryTree(LexEntry lex) {
		lexEntries.add(lex);
		convertLexEntry2ElementaryTree(lex);
		lex.setElementaryTree(this);
		set_is_modifier_auxiliary_tree();
	}
	

	public NodeStatus getType() {
		return type;
	}
	
	public void setType(NodeStatus ns) {
		type = ns;
	}
	
	public ElementaryTreeNode getRoot() {
		return root;
	}
	
	public void setRoot(ElementaryTreeNode r) {
		root = r;
	}
	
	public ElementaryTreeNode getFoot() {
		return foot;
	}
	
	public void setFoot(ElementaryTreeNode r) {
		foot = r;
	}
	
	public ElementaryTreeNode getAnchor() {
		return anchor;
	}
	
	public void setAnchor(ElementaryTreeNode r) {
		anchor = r;
	}
	
	public HashMap<Integer, ElementaryTreeNode> getNodes() {
		return nodes;
	}
	
	public void addNode(ElementaryTreeNode node) {
		nodes.put(node.getID(), node);
	}
	
	public void addNodes(HashMap n) {
		nodes.putAll(n);
	}
	
	
	public ElementaryTreeNode containsID(int id) {
//		if (nodes.containsKey(id)) {
//			return nodes.get(id);
//		}
//		else {
//			return null;
//		}
		return containsID(id, root);
	}
	
	private ElementaryTreeNode containsID(int id, ElementaryTreeNode node) {
		if (node.getID() == id) {
			return node;
		}
		else {
			ElementaryTreeNode found = null;
			for (ElementaryTreeNode child : node.getChildren()) {
				found = containsID(id, child);
				if (found != null) {
					break;
				}
			}
			return found;
		}
	}
	
	public ElementaryTreeNode getNodeWithTopAndBottom(TreeNode top, TreeNode bot) {
		return getNodeWithTopAndBottom(top, bot, root);
	}
	
	private ElementaryTreeNode getNodeWithTopAndBottom(TreeNode top, TreeNode bot, ElementaryTreeNode node) {
		if (node.getTreeNodeBottom() == bot 
				&& node.getTreeNodeTop() == top) {
			return node;
		}
		ElementaryTreeNode found = null;
		for (ElementaryTreeNode child : node.getChildren()) {
			found = getNodeWithTopAndBottom(top, bot, child);
			if (found != null) {
				break;
			}
		}
		return found;
	}
	
	public ElementaryTreeNode getNodeWithBottom(TreeNode bot) {
		return getNodeWithBottom(bot, root);
	}
	
	private ElementaryTreeNode getNodeWithBottom(TreeNode bot, ElementaryTreeNode node) {
		if (node.getTreeNodeBottom() == bot) {
			return node;
		}
		ElementaryTreeNode found = null;
		for (ElementaryTreeNode child : node.getChildren()) {
			found = getNodeWithBottom(bot, child);
			if (found != null) {
				break;
			}
		}
		return found;
	}
	
	public ElementaryTreeNode getNodeWithTop(TreeNode top) {
		return getNodeWithTop(top, root);
	}
	
	private ElementaryTreeNode getNodeWithTop(TreeNode top, ElementaryTreeNode node) {
		if (node.getTreeNodeTop() == top) {
			return node;
		}
		ElementaryTreeNode found = null;
		for (ElementaryTreeNode child : node.getChildren()) {
			found = getNodeWithTop(top, child);
			if (found != null) {
				break;
			}
		}
		return found;
	}
	
	public ElementaryTreeNode getNodeWithBottom(int id) {
		return getNodeWithBottom(id, root);
	}
	
	private ElementaryTreeNode getNodeWithBottom(int id, ElementaryTreeNode node) {
		if (node.getTreeNodeBottom().getID() == id) {
			return node;
		}
		ElementaryTreeNode found = null;
		for (ElementaryTreeNode child : node.getChildren()) {
			found = getNodeWithBottom(id, child);
			if (found != null) {
				break;
			}
		}
		return found;
	}
	
	public ElementaryTreeNode getNodeWithTop(int id) {
		return getNodeWithTop(id, root);
	}
	
	private ElementaryTreeNode getNodeWithTop(int id, ElementaryTreeNode node) {
		if (node.getTreeNodeTop().getID() == id) {
			return node;
		}
		ElementaryTreeNode found = null;
		for (ElementaryTreeNode child : node.getChildren()) {
			found = getNodeWithTop(id, child);
			if (found != null) {
				break;
			}
		}
		return found;
	}
	
	public TreeNode getModParent() {
		return modParent;
	}
	
	public void setModParent(TreeNode node) {
		modParent = node;
	}
	
	public boolean getMotherLevelForSisterAdj() {
		return motherLevelForSisterAdj;
	}
	
	public void setMotherLevelForSisterAdj(boolean b) {
		motherLevelForSisterAdj = b;
	}
	
	public HashMap<String, ElementaryTreeNode[]> getTraceSet() {
		return traceSet;
	}
	
	public boolean isRightAdj() {
		if (getType() != NodeStatus.ADJ) {
			return false;
		}
		// depth-first left-to-right traversal
		return traverseFindFoot(root);
	}
	

	private boolean traverseFindFoot(ElementaryTreeNode node) {
		if (node == foot) {
			return true;
		}
		else if (node.isSubst() || (node.isLeaf() && !node.isTrace())) {
			return false;
		}
		else {
			for (ElementaryTreeNode child : node.getChildren()) {
				return traverseFindFoot(child);
			}
		}
		return false;
	}
	
	public boolean isOnSpine(int id) {
		if (type != NodeStatus.ADJ) {
			return false;
		}
		if (id == foot.getID() || id == root.getID()) {
			return true;
		}
		ElementaryTreeNode cand = foot.getParent();
		while (cand != null) {
			if (id == cand.getID()) {
				return true;
			}
			cand = cand.getParent();
		}
		return false;
	}
	
	public boolean isLeftOfTrunk(ElementaryTreeNode node) {
		ElementaryTreeNode found = isLeftOfTrunkHelper(root, node);
		if (found == node) {
			return true;
		}
		return false;
	}
	
	private ElementaryTreeNode isLeftOfTrunkHelper(ElementaryTreeNode current,
			ElementaryTreeNode node) {
		if (current == node) {
			return node;
		}
		if (current == anchor) {
			return anchor;
		}
		ElementaryTreeNode found = null;
		for (ElementaryTreeNode child : current.getChildren()) {
			found = isLeftOfTrunkHelper(child, node);
			if (found != null) {
				break;
			}
		}
		return found;
	}

	public ElementaryTreeNode getLeftMostAnchor() {
		if (anchor == null) {
			return null;
		}
		if (getSecondaryAnchors().size() == 0) {
			return anchor;
		}
		// else we have to find out which of the anchors is the left-most one
		return getLeftMostAnchorHelper(root);
	}

	private ElementaryTreeNode getLeftMostAnchorHelper(ElementaryTreeNode node) {
		if (node == anchor) {
			return node;
		}
		for (ElementaryTreeNode secAnch : getSecondaryAnchors()) {
			if (secAnch.getID() == node.getID()) {
				return node;
			}
		}
		
		for (ElementaryTreeNode child : node.getChildren()) {
			ElementaryTreeNode anch = getLeftMostAnchorHelper(child);
			if (anch != null) {
				return anch;
			}
		}
		return null;
	}

	private void convertLexEntry2ElementaryTree(LexEntry lex) {
		// type of tree
		if (lex.isAdjTree()) {
			type = NodeStatus.ADJ;
		}
		else if (lex.getRoot().getNodeStatus() == NodeStatus.ARG ||
				lex.getRoot().getNodeStatus() == NodeStatus.MAARG) {
			type = NodeStatus.ARG;
		}
		else if (lex.getRoot().getNodeStatus() == NodeStatus.MOD) {
			type = NodeStatus.MOD;
			modParent = lex.getRoot().getParent();
		}
		else {
			System.err.println("What kind of tree is this? " + lex);
		}
		
		convertLexEntry2ElementaryTreeHelper(lex.getRoot(), lex);
		
		// set morphology (so far only case for NPs)
		//morphAnnotator.setMorphology(this);
		
	}

	private ElementaryTreeNode convertLexEntry2ElementaryTreeHelper(TreeNode node, LexEntry lex) {
		// Remember that node
		TreeNode original = node;
		
		// Skip adjunction trees
		while (node.getBottom() != node.getTop() && node.getBottom().isAdjTree() &&
				!node.getBottom().equals(lex) && !node.equals(lex.getFoot())) {
			node = node.getBottom().getFoot();
		}
		
		ElementaryTreeNode eTNode = new ElementaryTreeNode(node.getPhrasalCategory(), 
				original.getFunctionalCategory(), original, node, original.getID());
		
		if (original == lex.getRoot()) {
			root = eTNode;
			// Fix ID for adj and arg trees' root
			if (type == NodeStatus.ADJ || 
					(type == NodeStatus.ARG && lex.getRoot().getTop() != null)) { 
			//	eTNode.setID(node.getID());
			}
		}
		
		if (node.getBottom() != lex) {
			if (lex.isAdjTree() && node == lex.getFoot()) {
				foot = eTNode;
			}
			else if (original.getNodeStatus() == NodeStatus.ARG ||
					original.getNodeStatus() == NodeStatus.MAARG) {
				eTNode.setSubst();
			}
		}		
		else if (node.isLeaf()) {
			if (node.getBottom().getLeafLink() == node) {
				anchor = eTNode;
			}
			eTNode.setLeaf();
			String lf = ((LeafNode)node).getLeafString();
			String lm = ((LeafNode)node).getLemma();
			eTNode.setLeaf(lf);
			if (lm.matches("--")) {
				eTNode.setLemma(lf);
			}
			else {
				eTNode.setLemma(lm);
			}
			
			// trace
			if (((LeafNode)node).isTrace()) {
				eTNode.setTrace();
				treatTrace(eTNode, node.traceID);
			}
			
			eTNode.getMorphology().setSourceString(((LeafNode)node).getMorphology().getSourceString());
		}
		
		// filler
		if (original.isFiller() && original.getTop() != null && original.getTop().equals(lex)) {
			eTNode.setFiller();
			treatTrace(eTNode, original.traceID);
		}
		
		// set origins
		if (eTNode == getRoot() && (getType() == NodeStatus.ARG || getType() == NodeStatus.ADJ)) {
			eTNode.setOriginBottom(this);
		}
		else if (eTNode.isSubst() || eTNode == getFoot()) {
			eTNode.setOriginTop(this);
		}
		else {
			// inner nodes, leaf nodes, root nodes of MOD
			eTNode.setOriginTop(this);
			eTNode.setOriginBottom(this);
		}
		
		
		nodes.put(eTNode.getID(), eTNode);
		
		for (TreeNode child : node.getChildren()) {
			if (lex.containsNode(child)) {
				ElementaryTreeNode eTChild = convertLexEntry2ElementaryTreeHelper(child, lex);
				eTNode.addChild(eTChild);
				eTChild.setParent(eTNode);
				if (child.isHead()) {
					eTChild.setHead();
					eTNode.setHeadChild(eTChild);
				}
			}
		}
		
		return eTNode;
	}
	
	private void treatTrace(ElementaryTreeNode eTNode, String traceID) {
		ElementaryTreeNode[] trace;
		if (!traceSet.containsKey(traceID)) {
			trace = new ElementaryTreeNode[2];
		}
		else {
			trace = traceSet.get(traceID);
		}
		if (eTNode.isTrace()) {
			trace[0] = eTNode;
		}
		else if (eTNode.isFiller()) {
			trace[1] = eTNode;
		}
		traceSet.put(traceID, trace);
		eTNode.setTraceID(traceID);
	}
	
	private void addTrace(ElementaryTreeNode[] newTrace, String traceID)  {
		ElementaryTreeNode[] trace;
		if (!traceSet.containsKey(traceID)) {
			trace = new ElementaryTreeNode[2];
		}
		else {
			trace = traceSet.get(traceID);
		}
		if (newTrace[0] != null) {
			trace[0] = newTrace[0];
		}
		if (newTrace[1] != null) {
			trace[1] = newTrace[1]; // ?
		}
		traceSet.put(traceID, trace);
		
	}
	
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		toString(root, sb, 0);
		return sb.toString();
	}
	
	private void toString(ElementaryTreeNode node, StringBuffer sb, int level) {
		for (int i = 0; i < level; i++) {
			sb.append("  ");
		}
		sb.append(node.toString());
		sb.append("\n");
		
		for (ElementaryTreeNode child : node.getChildren()) {
			toString(child, sb, level+1);
		}
	}

	public String toStringLex(boolean motherLevelForSisterAdjunction,
			HashMap<String, Trace> traceListTigerTree) {
		setMotherLevelForSisterAdj(motherLevelForSisterAdjunction);
		return toStringLex(traceListTigerTree);
	}
	
	public String toStringLex(HashMap<String, Trace> traceListTigerTree) {
		StringBuilder sb = new StringBuilder();
		sb.append(anchor.getLeaf());
		sb.append(" ");
		sb.append(anchor.getLemma());
		for (ElementaryTreeNode secondary : getSecondaryAnchors()) {
			sb.append(" ").append(secondary.getLeaf()).append(" ").append(secondary.getLemma());
		}
		sb.append("\t");
		sb.append(type);
		sb.append("\t");
		sb.append(toStringStruct(traceListTigerTree));
		sb.append("\t");
		sb.append(anchor.getMorphology().getSourceString().toLowerCase());
		sb.append("\t");
		if (getAnchor().isTrace()) {
			sb.append("EMPTY-ANCHOR");
		}
		return sb.toString();
	}
	
	
	public String toStringStruct(HashMap<String, Trace> traceListTigerTree) {
		StringBuilder sb = new StringBuilder();
		
		traceMappingForOutput.clear();
		
		// add the mother level to modifier trees 
		// if the motherLevelForSisterAdj option is set accordingly
		if (motherLevelForSisterAdj && type == NodeStatus.MOD) {
			int top = individuals.indexOf(root.getOriginTop());
			String topOrigin;
			if (top == -1) {
				if (root.getOriginTop() == this) topOrigin = "x";
				else topOrigin = "null";
			}
			else {
				topOrigin = String.valueOf(top+1);
			}
			
			// TODO fix functional category for the case that an auxiliary
			// tree was adjoined into the modparent, e.g. s600
			sb.append("(");
			sb.append(getModParent().getPhrasalCategory());
			if (!getModParent().getFunctionalCategory().equals("--")) {
				sb.append("-");
				sb.append(getModParent().getFunctionalCategory());
			}
			sb.append("^null_").append(topOrigin);
			sb.append("* ");
			
			toStringStruct(root, sb, traceListTigerTree);
			
			sb.append(")");
		}
		else {
			toStringStruct(root, sb, traceListTigerTree);
		}
		return sb.toString();
	}

	private void toStringStruct(ElementaryTreeNode node, StringBuilder sb,
			HashMap<String, Trace> traceListTigerTree) {
		
		int top = individuals.indexOf(node.getOriginTop());
		int bottom = individuals.indexOf(node.getOriginBottom());
		String topOrigin;
		String bottomOrigin;
		if (top == -1) {
			if (node.getOriginTop() == this) topOrigin = "x";
			else topOrigin = "null";
		}
		else {
			topOrigin = String.valueOf(top+1);
		}
		if (bottom == -1) {
			if (node.getOriginBottom() == this) bottomOrigin = "x";
			else bottomOrigin = "null";
		}
		else {
			bottomOrigin = String.valueOf(bottom+1);
		}
		
//		// mark foot node of modifier auxiliary tree
//		if (type == NodeStatus.ADJ && foot == node && is_modifier_auxiliary_tree()) {
//			topOrigin += "M";
//		}
//		// mark coordinated categories
//		if (node.getTreeNodeBottom().isCoordinatedCat()) {
//			bottomOrigin += "C";
//		}
//		if (node.getTreeNodeBottom().isUnlikeCoordination()) {
//			bottomOrigin += "O";
//		}
//		// mark nodes that had a trace child (that is has been deleted)
//		if (node.getTreeNodeBottom().hadTraceChild()) {
//			bottomOrigin += "E";
//		}
//		// mark nodes that have incoming secondary edges
//		if (node.getTreeNodeBottom().getIncomingSecondaryEdges().size() > 0) {
//			bottomOrigin += "S";
//		}
		// CHANGE
		// mark nodes that have been inserted for binarization
		//if (node.getTreeNodeBottom().insertedForBinarization()) {
		//	bottomOrigin += "B";
		//}
		//if (node.getTreeNodeTop().insertedForBinarization()) {
		//	topOrigin += "B";
		//}
		
		sb.append("(");
		
		if (node.isTrace()) {
			if (!traceMappingForOutput.containsKey(node.getTraceID())) {
				traceMappingForOutput.put(node.getTraceID(), traceMappingForOutput.size()+1);
			}
			sb.append("*T");
			sb.append(traceMappingForOutput.get(node.getTraceID()));
			sb.append("*");
		}
		else {
		
			// phrasal category
			sb.append(node.getPhrasalCategory());
		}
		
		// moved further down
//		// filler information 
//		if (node.isFiller() && traceListTigerTree.containsKey(node.getTraceID())) {
//			// CHANGE
//			//sb.append("-");
//			//sb.append(node.getTraceID());
//			
//			if (!traceMappingForOutput.containsKey(node.getTraceID())) {
//				traceMappingForOutput.put(node.getTraceID(), traceMappingForOutput.size()+1);
//			}
//			sb.append("{T");
//			sb.append(traceMappingForOutput.get(node.getTraceID()));
//			sb.append("}");
//		}
		
		// functional category
		if (!node.getFunctionalCategory().equals("--")) {
			sb.append("-");
			sb.append(node.getFunctionalCategory());
		}
		
		// morphology
		sb.append(node.getMorphology());
		
		// filler information
		if (node.isFiller() && traceListTigerTree.containsKey(node.getTraceID())) {
			// CHANGE
			//sb.append("-");
			//sb.append(node.getTraceID());
			
			if (!traceMappingForOutput.containsKey(node.getTraceID())) {
				traceMappingForOutput.put(node.getTraceID(), traceMappingForOutput.size()+1);
			}
			sb.append("{T");
			sb.append(traceMappingForOutput.get(node.getTraceID()));
			sb.append("}");
		}
		
		// secondary edge information
		if (!bottomOrigin.equals("null")) {
			for (SecondaryEdge sec : node.getTreeNodeBottom().getIncomingSecondaryEdges()) {
				TreeNode dependent = sec.getStartNode();
				NodeStatus ns = dependent.getNodeStatus();
				if (ns == NodeStatus.ARG || ns == NodeStatus.MAARG || dependent.isHead() 
						|| (ns == NodeStatus.MOD && Utils.isAuxiliaryOrModal(dependent))) {
					sb.append("$$").append(dependent.getPhrasalCategory());
					if (!dependent.getFunctionalCategory().equals("--")) {
						sb.append("-").append(dependent.getFunctionalCategory());
					}
					sb.append(dependent.getTop().getElementaryTree().getNodeWithTop(dependent).getMorphology());
					sb.append("$$");
				}
				//else {
				//	System.out.println("MODifier");
				//}
			}
		}
		
		if (node == root) {
			if (type == NodeStatus.MOD) {
				sb.append("^").append(topOrigin).append("_").append(bottomOrigin);
				if (!motherLevelForSisterAdj) {
					sb.append("*");
				}
			}
			else if (type == NodeStatus.ARG || type == NodeStatus.ADJ) {
				sb.append("^").append("null").append("_").append(bottomOrigin);
			}
		}
		else if (node.isSubst()) {
			sb.append("^").append(topOrigin).append("_").append("null");
			sb.append("!");
		}
		else if (type == NodeStatus.ADJ && foot == node) {
			sb.append("^").append(topOrigin).append("_").append("null");
			sb.append("*");
		}
		else {
			sb.append("^").append(topOrigin).append("_").append(bottomOrigin);
		}
		
		sb.append(" ");
		if (node.isLeaf()) {
			sb.append(node.getLeaf());
			if (node == anchor) {
				sb.append("<>");
			}
			else if(!node.isTrace()) {
				sb.append("^").append(bottomOrigin).append("_").append(bottomOrigin);
				// (bottomOrigin because it is supposed to be the same index as its POS
				// which is the same node...)
			}
		}
		else {
			for (ElementaryTreeNode child : node.getChildren()) {
				toStringStruct(child, sb, traceListTigerTree);
			}
		}
		
		sb.append(")");
		
	}
	
	/*
	 * Find all "open" substitution nodes in this elementary tree
	 */
	public ArrayList<ElementaryTreeNode> getSubstitutionSites() {
		ArrayList<ElementaryTreeNode> substs = new ArrayList<ElementaryTreeNode>();
		getSubstitutionSites(root, substs);
		return substs;
	}

	private void getSubstitutionSites(ElementaryTreeNode node, ArrayList<ElementaryTreeNode> substs) {
		if (node.isSubst() && node.getChildren().size() == 0) {
			substs.add(node);
		}
		
		for (ElementaryTreeNode child : node.getChildren()) {
			getSubstitutionSites(child, substs);
		}
	}
	
	public boolean isPartOfMAT() {
		return partOfMAT;
	}
	
	public void setPartOfMAT() {
		partOfMAT = true;
	}
	
	public ArrayList<ElementaryTree> getIndividualElementaryTrees() {
		return individuals;
	}
	
	public void addIndividualElementarytree(ElementaryTree e) {
		individuals.add(e);
	}
	
	public ArrayList<ElementaryTreeNode> getSecondaryAnchors() {
		return otherAnchors;
	}
	
	public void addSecondaryAnchor(ElementaryTreeNode node) {
		otherAnchors.add(node);
	}

	public boolean turnIntoMultiAnchoredTree(boolean mainAnchorAnywhereInMAT) {
		boolean success = true;
		ArrayList<LexEntry> others = lexEntries.get(0).getAssociates();
		for (LexEntry other : others) {
			ElementaryTree otherETree = other.getElementaryTree();
			success = combine(otherETree, mainAnchorAnywhereInMAT);
			
			if (!success) return success;
			
			lexEntries.add(other);
			otherETree.setPartOfMAT();
			addIndividualElementarytree(otherETree);	
			addSecondaryAnchor(otherETree.getAnchor());
		}
		return success;
	}

	private boolean combine(ElementaryTree eTree, boolean mainAnchorAnywhereInMAT) {
		boolean integrated = false;
		if (eTree.getType() == NodeStatus.ARG) {
			ElementaryTreeNode node = getNodeWithBottom(eTree.getRoot().getTreeNodeBottom());
			if (node != null && node.isSubst()) {
				if ( node.getChildren().size() == 0 &&
						node.getPhrasalCategory().equals(eTree.getRoot().getPhrasalCategory())) {
					
					if (!mainAnchorAnywhereInMAT && isLeftOfTrunk(node)) {
						// If the left-most anchor has to be the main anchor,
						// we have to skip some trees.
						// The reason is that when calculation the prediction trees later, the assumption
						// will be that the left-most anchor is the point of recognition.
						return false;
					}
					
					ElementaryTreeNode parent = node.getParent();
					// replace this substitution site with the initial eTree
					int i = parent.getChildren().indexOf(node);
					getNodes().remove(node.getID());
					ElementaryTree eTreeCopy = makeCopy(eTree);
					addNodes(eTreeCopy.getNodes());
					// keep the "top" part of the substitution node
					eTreeCopy.getRoot().setFunctionalCategory(node.getFunctionalCategory());
					eTreeCopy.getRoot().setOriginTop(node.getOriginTop());
					eTreeCopy.getRoot().setTreeNodeTop(node.getTreeNodeTop());
					parent.setChild(i, eTreeCopy.getRoot());
					eTreeCopy.getRoot().setParent(parent);
					// traces
					for (String id : eTreeCopy.getTraceSet().keySet()) {
						addTrace(eTreeCopy.getTraceSet().get(id), id);
					}
					if (node.isFiller()) {
						eTreeCopy.getRoot().setFiller();
						eTreeCopy.getRoot().setTraceID(node.getTraceID());
					}
					integrated = true;
				}
			}
		}
		return integrated;
	}

	private ElementaryTree makeCopy(ElementaryTree eTree) {
		ElementaryTree newETree = new ElementaryTree();
		newETree.setType(eTree.getType());
		makeCopyHelper(eTree.getRoot(), eTree, newETree);
		if (eTree.getModParent() != null) {
			newETree.setModParent(eTree.getModParent());
		}
		traceSet.putAll(eTree.getTraceSet());
		return newETree;
	}
	
	private ElementaryTreeNode makeCopyHelper(ElementaryTreeNode node, ElementaryTree eTree,
					ElementaryTree newETree) {
		
		ElementaryTreeNode copy = new ElementaryTreeNode(node);
		
		if (eTree.getType() == NodeStatus.ADJ && node == eTree.getFoot()) {
			newETree.setFoot(copy);
		}
		if (node == eTree.getRoot()) {
			newETree.setRoot(copy);
		}
		if (node == eTree.getAnchor()) {
			newETree.setAnchor(copy);
		}
		
		for (ElementaryTreeNode child : node.getChildren()) {
			ElementaryTreeNode childCopy = makeCopyHelper(child, eTree, newETree);
			childCopy.setParent(copy);
			copy.addChild(childCopy);
			if (child.isHead()) {
				childCopy.setHead();
				copy.setHeadChild(childCopy);
			}
		}
		
		newETree.addNode(copy);
		return copy;
	}

	public String calcGornTreeNodeBottom(int id) {
		return calcGornTreeNodeBottom(id, root, "");
	}

	private String calcGornTreeNodeBottom(int id, ElementaryTreeNode node, String gorn) {
		if (node.getTreeNodeBottom().getID() == id) {
			if (gorn == "") return "0";
			else return gorn;
		}
		for (int i = 0; i < node.getChildren().size(); i++) {
			int j = i+1;
			String newGorn = gorn + j;
			String g = calcGornTreeNodeBottom(id, node.getChildren().get(i), newGorn);
			if (!g.equals("null")) return g;
		}
		return "null";
	}

	public String calcGornTreeNodeTop(int id) {
		return calcGornTreeNodeTop(id, root, "");
	}

	private String calcGornTreeNodeTop(int id, ElementaryTreeNode node,
			String gorn) {
		if (node.getTreeNodeTop().getID() == id) {
			if (gorn == "") return "0";
			else return gorn;
		}
		for (int i = 0; i < node.getChildren().size(); i++) {
			int j = i+1;
			String newGorn = gorn + j;
			String g = calcGornTreeNodeTop(id, node.getChildren().get(i), newGorn);
			if (!g.equals("null")) return g;
		}
		return "null";
	}

	private void set_is_modifier_auxiliary_tree() {
		boolean footFound = false;
		boolean modFound = false;
		if (type != NodeStatus.ADJ) {
			return;
		}
		for (ElementaryTreeNode child : root.getChildren()) {
			if (child == foot) {
				footFound = true;
			}
			if (child.getTreeNodeTop().getNodeStatus() == NodeStatus.MOD) {
				modFound = true;
			}
		}
		if (modFound && footFound) {
			modifier_auxiliary = true;
		}
		
	}
	
	public boolean is_modifier_auxiliary_tree() {
		return modifier_auxiliary;
	}
}
