package pltag.german;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;


public class HeadPercolTable {
	
	private HashMap<String, ArrayList<String>> map = new HashMap<String, ArrayList<String>>();
	private HashMap<String, Boolean> fromLeft = new HashMap<String, Boolean>();
	
	private int count;
	
	public HeadPercolTable(File filename) {
		read(filename);		
		//printPt();
		count = 0;
	}
	
	public int getCount() {
		return count;
	}
	
	public void printPt(){
		for (Map.Entry<String, ArrayList<String>> entry : map.entrySet()) {
			System.out.println(entry.getKey() + " " + 
					fromLeft.get(entry.getKey()) + " " + entry.getValue());
		}
	}
	
	public void setHeadsInTree(TreeNode node) {
		setHead(node);
		for (TreeNode child : node.getChildren()) {
			setHeadsInTree(child);
		}
	}
	
	public boolean setHead(TreeNode node) {
		count++;
		String cat = node.getPhrasalCategory();
		ArrayList<TreeNode> children = node.getChildren();
		boolean headFound = false;
		if (fromLeft.get(cat)) {
			outerLoop: for (String potHeadCat : map.get(cat)) {
				for (TreeNode child : children) {
					if (child.getPhrasalCategory().equals(potHeadCat)) {
						node.setHeadChild(child);
						child.setHead();
						headFound = true;
						break outerLoop;
					}
				}
			}
		}
		else {
			outerLoop: for (String potHeadCat : map.get(cat)) { 
				for (int i = children.size()-1; i > -1 ; i--) {
				TreeNode child = children.get(i);
					if (child.getPhrasalCategory().equals(potHeadCat)) {
						node.setHeadChild(child);
						child.setHead();
						headFound = true;
						break outerLoop;
					}
				}
			}
		}
		return headFound;
	}
	

	private void read(File filename) {
		BufferedReader input = null;
		try { //FileReader always assumes default encoding is OK!
			input = new BufferedReader( new FileReader(filename) );
			String line = null; 
			while (( line = input.readLine()) != null){
				String[] list = line.split("\\s+");
				ArrayList<String> categories = new ArrayList<String>();
				Boolean direction_fromleft = false;
				if (list[1].equals("left")){direction_fromleft = true;}
				
				String key = list[0];
				for (int i = 2; i < list.length; i++){
					categories.add(list[i]);
				}
				
				map.put(key, categories);				
				fromLeft.put(key, direction_fromleft);
			}	
		}
		catch (FileNotFoundException ex) {ex.printStackTrace();}
		catch (IOException ex){ex.printStackTrace();}
		finally {
			try {if (input!= null) {input.close();}}
			catch (IOException ex) {ex.printStackTrace();}
		}
	}


}
