package pltag.german.analysis;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class ModTreesSubstOfInitTrees {

	private int inInit = 0;
	private int modAll = 0;
	private HashSet<String> initialTreeTemplates = new HashSet<String>();
	
	private Pattern pOrigin = Pattern.compile("(\\-\\*T1\\*)?\\^[x\\d]_([x\\d])\\*");

	
	public static void main(String[] args) {
		
		// the output files from Analysis.java
		// note that this was written for the "old" modifier trees (without additional level)
		File ana_mod = new File(args[0]);
		File ana_init = new File(args[1]);
		
		ModTreesSubstOfInitTrees obj = new ModTreesSubstOfInitTrees(ana_init);
		obj.analyze(ana_mod);
	}


	public ModTreesSubstOfInitTrees(File init) {
		// read in all the initial trees
		readInitialTrees(init);
	}

	private void readInitialTrees(File init) {
		FileReader fr ;
		BufferedReader input = null;
		
		try {			
			fr = new FileReader(init);
			input = new BufferedReader(fr);
			
			String line = null;	
			while (( line = input.readLine()) != null){
				if (line.startsWith("%%") || line.startsWith("\t")) { // comment or lexical item
					continue;
				}
				
				if (line.startsWith("(")) { // tree template
					initialTreeTemplates.add(line);
					
				}				
			}
		}
	
			
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
		catch (IOException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				if (input!= null) {
					input.close();
				}
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}

	private void analyze(File ana_mod) {
		inInit = 0;
		modAll = 0;
		
		
		FileReader fr ;
		BufferedReader input = null;
		
		try {			
			fr = new FileReader(ana_mod);
			input = new BufferedReader(fr);
			
			String line = null;	
			while (( line = input.readLine()) != null){
				if (line.startsWith("%%") || line.startsWith("\t")) { // comment or lexical item
					continue;
				}
				String tree = "";
				if (line.startsWith("(")) { // tree template
					Matcher m1 = pOrigin.matcher(line);
					if(m1.find()) {
						tree = m1.replaceFirst("^null_"+m1.group(2));
					}
					else {
						System.err.println("Error in root of " + line);
						continue;
					}
					
					modAll++;
					if (initialTreeTemplates.contains(tree)) {
						inInit++;
					}
					else {
						System.out.println(line);
					}
				}				
			}
		}
		catch (FileNotFoundException e) {
			System.err.println(e.getMessage());
			//e.printStackTrace();
		}
		catch (IOException ex){
			ex.printStackTrace();
		}
		finally {
			try {
				if (input!= null) {
					input.close();
				}
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
		
		System.out.println("Number of modifier trees: " + modAll);
		System.out.println("Number of initial trees: " + initialTreeTemplates.size());
		System.out.println("Number of modifier trees in initial trees: " + inInit);
		
	}
}
