package pltag.german.regularadjunction;

import java.util.ArrayList;

import pltag.german.InnerNode;
import pltag.german.NodeStatus;
import pltag.german.TigerTree;
import pltag.german.TreeNode;
import pltag.german.Utils;

public class HeadOutwardBinarizer {
	
	private static final String HD = "HD";
	
	private TigerTree currTree;
	
	public HeadOutwardBinarizer() {
		
	}

	public void binarize(TigerTree tree) {
		currTree = tree;
		binarizeLevel(tree.getRoot());
		
	}

	private void binarizeLevel(TreeNode node) {
		if (node.isLeaf()) return;
		
		if (node.getPhrasalCategory().equals("S")) {
			node = treatSentenceFinalPunctuation(node);
		}
		
		TreeNode leftChild = node.getFirstChild();
		while (!leftChild.isHead()) {
			binarizeLevel(leftChild);
			if (node.getChildren().size() < 3 
					&& node.getHeadChild().getPhrasalCategory().equals(node.getPhrasalCategory())) {
				break;
			}
			node = rightbinarize(node, leftChild);
			leftChild = node.getFirstChild();
		}
		
		TreeNode rightChild = node.getLastChild();
		while(!rightChild.isHead()) {
			binarizeLevel(rightChild);
			if (node.getChildren().size() < 3 
					&& node.getHeadChild().getPhrasalCategory().equals(node.getPhrasalCategory())) {
				break;
			}
			node = leftbinarize(node, rightChild);
			rightChild = node.getLastChild();
		}
		
		binarizeLevel(node.getHeadChild());
	}


	private TreeNode rightbinarize(TreeNode node, TreeNode child) {		
		TreeNode hd = node.getHeadChild();
		String newCat = findPhrasalCatForNewNode(node, child);	
		
		// special case
		if (newCat.equals(hd.getPhrasalCategory()) 
				&& node.getChildren().size() < 3 ) {
			return hd;
		}
		
		TreeNode newNode = new InnerNode(currTree.getIDGenerator(), 
				newCat + "-" + HD, true);
		currTree.getID2Node().put(newNode.getID(), newNode);
		
		int ind = node.getChildren().indexOf(child);
		for (int i = node.getChildren().size()-1; i > ind; i--) {
			TreeNode c = node.getChildren().get(i);
			node.removeChild(i);
			newNode.addChild(0, c);
			c.setParent(newNode);
		}
		node.addChild(ind+1, newNode);
		newNode.setParent(node);
		
		// fix heads
		node.setHeadChild(newNode);
		newNode.setHead();
		newNode.setHeadChild(hd);
		
		// fix identification of predicative auxiliary trees
		if (node.getNodeStatusBottom() == NodeStatus.PREDADJ
				&& node.getPhrasalCategory().equals(newNode.getPhrasalCategory())
				&& node.getPhrasalCategory().matches("S|VP")
				&& child.getNodeStatus() == NodeStatus.MOD) {
			TreeNode foot = currTree.getRoot2Foot().get(node.getID());
			currTree.getRoot2Foot().remove(node.getID());
			currTree.getRoot2Foot().put(newNode.getID(), foot);
			currTree.getFoot2Root().put(foot.getID(), newNode);
			node.setNodeStatusBottom(NodeStatus.UNDEF);
			newNode.setNodeStatusBottom(NodeStatus.PREDADJ);
		}
		
		// fix identification of coordination auxiliary trees
		if (node.isCoordinatedCat()
				&& child.getNodeStatus() == NodeStatus.MOD) {
			newNode.setCoordinatedCat(true);
			node.setCoordinatedCat(false);
		}
		
		return newNode;
	}
	

	private TreeNode leftbinarize(TreeNode node, TreeNode child) {		
		TreeNode hd = node.getHeadChild();
		String newCat = findPhrasalCatForNewNode(node, child);
		TreeNode newNode = new InnerNode(currTree.getIDGenerator(), 
				newCat + "-" + HD, true);
		currTree.getID2Node().put(newNode.getID(), newNode);
		
		int ind = node.getChildren().indexOf(child);
		for (int i = ind-1; i > -1; i--) {
			TreeNode c = node.getChildren().get(i);
			node.removeChild(i);
			newNode.addChild(0, c);
			c.setParent(newNode);
		}
		node.addChild(0, newNode);
		newNode.setParent(node);
		
		// fix heads
		node.setHeadChild(newNode);
		newNode.setHead();
		newNode.setHeadChild(hd);
		
		// fix identification of predicative auxiliary trees
		if (node.getNodeStatusBottom() == NodeStatus.PREDADJ
				&& node.getPhrasalCategory().equals(newNode.getPhrasalCategory())
				&& node.getPhrasalCategory().matches("S|VP")
				&& child.getNodeStatus() == NodeStatus.MOD) {
			TreeNode foot = currTree.getRoot2Foot().get(node.getID());
			currTree.getRoot2Foot().remove(node.getID());
			currTree.getRoot2Foot().put(newNode.getID(), foot);
			currTree.getFoot2Root().put(foot.getID(), newNode);
			node.setNodeStatusBottom(NodeStatus.UNDEF);
			newNode.setNodeStatusBottom(NodeStatus.PREDADJ);
		}
		
		// fix identification of coordination auxiliary trees
		if (node.isCoordinatedCat()
				&& child.getNodeStatus() == NodeStatus.MOD) {
			newNode.setCoordinatedCat(true);
			node.setCoordinatedCat(false);
		}

		return newNode;
	}

	private String findPhrasalCatForNewNode(TreeNode node, TreeNode child) {
		// CHANGE
//		if (node.getPhrasalCategory().equals("S") && child.getNodeStatus() != NodeStatus.MOD
//				&& hasVerbalHeadChild(node)) {
//			return "VP";
//		}
		return node.getPhrasalCategory();
	}
	
	private boolean hasVerbalHeadChild(TreeNode node) {
		if (node.getPhrasalCategory().matches("(VV|VA|VM).*|VZ|CVP|VP")
				|| node.getFunctionalCategory().matches("SVP")) {
			return true;
		}
		else if (node.getHeadChild() != null){
			return hasVerbalHeadChild(node.getHeadChild());
		}
		return false;
	}
	
	private TreeNode treatSentenceFinalPunctuation(TreeNode node) {
		ArrayList<TreeNode> punct = new ArrayList<TreeNode>();
		for (int i = node.getChildren().size()-1; i >= 0; i--) {
			TreeNode child = node.getChildren().get(i);
			if (Utils.isPunct(child.getPhrasalCategory())) {
				punct.add(0, child);
			}
			else {
				break;
			}
		}
		
		if (punct.size() > 0) {
			TreeNode newNode = new InnerNode(currTree.getIDGenerator(), 
					"S" + "-" + HD, true);
			currTree.getID2Node().put(newNode.getID(), newNode);
			
			if (currTree.getRoot() == node) {
				currTree.setRoot(newNode);
			}
			else {
				TreeNode pa = node.getParent();
				int ind = pa.indexOfChild(node);
				pa.removeChild(ind);
				pa.addChild(ind, newNode);
				newNode.setParent(pa);
				if (node.isHead()) {
					newNode.setHead();
					pa.setHeadChild(newNode);
				}
			}
			NodeStatus ns = node.getNodeStatus();
			newNode.setNodeStatus(ns);
			newNode.addChild(node);
			node.setParent(newNode);
			newNode.setHeadChild(node);
			node.setHead();
			for (TreeNode p : punct) {
				node.removeChild(p);
				newNode.addChild(p);
				p.setParent(newNode);
			}
			return newNode;
		}
		return node;
	}

	
}
